/*****************************************************************************
 * Font support library                                                      *
 *                                                                           *
 * This is using the Sprite library to implement fonts, so you should become *
 * familiar with that first!                                                 *
 *****************************************************************************/
#include "Fonts.h"

// We will keep track of all the fonts loaded, so at uninitialization time,
// we can free the ones that are forgotten to be freed.
typedef struct _tLoadedFontList
{
  pFont Font;
  void *Next;
} tLoadedFontList, *pLoadedFontList;

pLoadedFontList LoadedFontList = NULL;

//---------------------------------------------------------------------
// LoadFontDef
//
// Loads the image of a character of a given font. Also allocates space
// for the font, if it was not allocated yet.
//---------------------------------------------------------------------
void LoadFontDef(pFont *pTheFont, pImage Image, char *Params)
{
  char *ptemp1;
  unsigned char *src, *dst;
  unsigned char *srcM, *dstM;
  pSprite *TheSprite;
  long xsize, ysize, xpos, ypos, mxpos, mypos, hxpos, hypos, y;
  int rc;
  pLoadedFontList LoadedFont;

  if ((!Params) || (!Image) || (!pTheFont)) return;

  // Allocate memory for the parameter parsing
  ptemp1=(char *) malloc(2048);
  if (!ptemp1) return;

  // Parse the parameters: <character> <xsize> <ysize> ....
  rc=sscanf(Params, "%s %d %d %d %d %d %d %d %d",
            ptemp1, &xsize, &ysize, &xpos, &ypos, &mxpos, &mypos,
            &hxpos, &hypos);
  if (rc!=9)
  {
    free(ptemp1);
    return;
  }

  // Ok, we have the parameters!
  // Let's see if it's the first fontdef for this font!
  // (because we have to allocate the space then...)
  if (!(*pTheFont))
  {
    *pTheFont = (pFont) malloc(sizeof(tFont));
    if (!(*pTheFont))
    {
      free(ptemp1);
      return;
    }
    memset((*pTheFont)->Character, 0, sizeof((*pTheFont)->Character));
#ifdef FONTS_DEBUG_BUILD
    printf("LoadFontDef -> NewFont!\n");
    fflush(stdout);
#endif

    // Put it into the list of loaded fonts
    LoadedFont = (pLoadedFontList) malloc(sizeof(tLoadedFontList));
    if (LoadedFont)
    {
      LoadedFont->Font = *pTheFont;
      LoadedFont->Next = LoadedFontList;
    }
  }

  // Get a pointer to the sprite of the character to be defined
  if (!stricmp(ptemp1, "SPACE"))
  {
    TheSprite=&((*pTheFont)->Character[' ']); // Space
  }
  else
  {
    TheSprite=&((*pTheFont)->Character[ptemp1[0]]); // Not space
  }

  // Cleanup the sprite, if it was already defined
  if (*TheSprite) CleanupSprite(*TheSprite);

  // And create a new sprite by hand
  *TheSprite = (pSprite) malloc(sizeof(tSprite));
  if (!(*TheSprite))
  {
    free(ptemp1);
    return;
  }
  // fill with the new data!
  (*TheSprite)->flAnimSpeed = 1;
  (*TheSprite)->ulActFrame = 0;
  (*TheSprite)->flActFrame = 0;
  (*TheSprite)->ulNumOfFrames = 1;  // Characters are not animated
  (*TheSprite)->ulOneFrameSize = xsize*ysize*3;
  (*TheSprite)->ulSize = (*TheSprite)->ulOneFrameSize * (*TheSprite)->ulNumOfFrames;
  (*TheSprite)->ulWidth = xsize;
  (*TheSprite)->ulHeight = ysize;
  (*TheSprite)->ulHotSpotX = hxpos;
  (*TheSprite)->ulHotSpotY = hypos;
  (*TheSprite)->pColorData = (unsigned char *)malloc((*TheSprite)->ulSize);
  if (!((*TheSprite)->pColorData))
  {
    free(*TheSprite); *TheSprite = NULL;
    free(ptemp1);
    return;
  }

  (*TheSprite)->pColorMask = (unsigned char *)malloc((*TheSprite)->ulSize);
  if (!((*TheSprite)->pColorMask))
  {
    free((*TheSprite)->pColorData);
    free(*TheSprite); *TheSprite = NULL;
    free(ptemp1);
    return;
  }

  // Copy the image and mask into the sprite
  src=Image->pImageData + Image->ulWidth*ypos*3 + xpos*3;
  dst=(*TheSprite)->pColorData;
  srcM=Image->pImageData + Image->ulWidth*mypos*3 + mxpos*3;
  dstM=(*TheSprite)->pColorMask;
  for (y=0; y<(*TheSprite)->ulHeight; y++)
  {
    memcpy(dst, src, (*TheSprite)->ulWidth*3);
    dst+=(*TheSprite)->ulWidth*3;
    src+=Image->ulWidth*3;
    memcpy(dstM, srcM, (*TheSprite)->ulWidth*3);
    dstM+=(*TheSprite)->ulWidth*3;
    srcM+=Image->ulWidth*3;
  }

  // That's all, we're ready!
  free(ptemp1);
}

//---------------------------------------------------------------------
// CleanupFont
//
// Cleans up a given font. It frees all memory allocated for it, but
// its your job to set the font handle to NULL, if you want to!
//---------------------------------------------------------------------
void CleanupFont(pFont Font)
{
  pLoadedFontList LoadedFont, prev;
  int i;

  if (!Font) return;

#ifdef FONTS_DEBUG_BUILD
  printf("CleanupFont()\n");
  fflush(stdout);
#endif

  // Search for the font in the list of loaded fonts
  prev = NULL;
  LoadedFont = LoadedFontList;
  while ((LoadedFont) && (LoadedFont->Font != Font))
  {
    prev = LoadedFont;
    LoadedFont = LoadedFont->Next;
  }
  // If found, remove from the list!
  if (LoadedFont)
  {
    if (prev)
      prev->Next = LoadedFont->Next;
    else
      LoadedFontList = LoadedFont->Next;
    free(LoadedFont);
  }

  // And now free every resource allocated for this font
  for (i=0; i<256; i++)
    if (Font->Character[i])
    { // Free every loaded character sprite
      CleanupSprite(Font->Character[i]);
      Font->Character[i] = NULL;
    }
  // and free the font descriptor itself
  free(Font);
}

//---------------------------------------------------------------------
// SetFontHeight
//
// Sets the height of a given font. This is used in case there is a
// multiline message to be shown. The next line will be this much under
// the previous one.
//---------------------------------------------------------------------
void SetFontHeight(pFont *pTheFont, int iHeight)
{
  pLoadedFontList LoadedFont;

  if (!pTheFont) return;

  // Let's see if it's the first fontdef for this font!
  // (because we have to allocate the space then...)
  if (!(*pTheFont))
  {
    *pTheFont = (pFont) malloc(sizeof(tFont));
    if (!(*pTheFont))
      return;
    memset((*pTheFont)->Character, 0, sizeof((*pTheFont)->Character));
#ifdef FONTS_DEBUG_BUILD
    printf("SetFontHeight -> NewFont!\n");
    fflush(stdout);
#endif
    // Insert into the list of loaded fonts!
    LoadedFont = (pLoadedFontList) malloc(sizeof(tLoadedFontList));
    if (LoadedFont)
    {
      LoadedFont->Font = *pTheFont;
      LoadedFont->Next = LoadedFontList;
    }
  }
  // And store this height with the font
  (*pTheFont)->iHeight = iHeight;
}

//---------------------------------------------------------------------
// DrawText
//
// Draws a given text message into a given BGR3 output buffer, using
// a give font.
// It is not possible to draw multiline text with this function!
//
// Parameters: 
//  pBuffer        - Pointer to image buffer
//  ulDstBufWidth  - Width of buffer in bytes (height is not important)
//  vptlx          - Viewport, top left corner, X coordinate
//  vptly          - Viewport, top left corner, Y coordinate
//  vpbrx          - Viewport, bottom right corner, X coordinate
//  vpbry          - Viewport, bottom right corner, Y coordinate
//  xpos, ypos     - Top-left corner of the text to be displayed
//  text           - The text to be displayed
//  TheFont        - The font to be used
//  Visibility     - Transparency of text (0..256, 0 = not visible)
//
// The text to output will be clipped into viewport
//---------------------------------------------------------------------
void DrawText(unsigned char *pBuffer, unsigned long ulDstBufWidth,
	      long vptlx, long vptly, long vpbrx, long vpbry,
	      long xpos, long ypos,
	      char *text,
              pFont TheFont, unsigned char Visibility)
{
  pSprite TheSprite;

  if ((!text) || (!pBuffer) || (!TheFont)) return;
  // Go through all the characters, and show them!
  while (*text)
  {
    TheSprite = TheFont->Character[*text];
    if (TheSprite)
    {
      DrawSprite(pBuffer, ulDstBufWidth, vptlx, vptly, vpbrx, vpbry, xpos, ypos, TheSprite, Visibility);
      xpos+=TheSprite->ulWidth - TheSprite->ulHotSpotX;
    }
    text++;
  }
}


//---------------------------------------------------------------------
// InitializeFonts()
//
// Initialize font support. Sprite support should be already 
// initialized tobe able to use fonts!
//---------------------------------------------------------------------
void InitializeFonts()
{
#ifdef FONTS_DEBUG_BUILD
  printf("InitializeFonts()\n");
  fflush(stdout);
#endif
  LoadedFontList = NULL;
}

//---------------------------------------------------------------------
// UninitializeFonts()
//
// Uninitializes font support. Frees all the fonts that are forgotten 
// to be freed.
//---------------------------------------------------------------------
void UninitializeFonts()
{
#ifdef FONTS_DEBUG_BUILD
  printf("UninitializeFonts()\n");
  fflush(stdout);
#endif

  while (LoadedFontList)
    CleanupFont(LoadedFontList->Font);
}
