/*****************************************************************************
 * Frame regulation library                                                  *
 *                                                                           *
 * Can be used to post a semaphore in a given number of times per second,    *
 * giving ticks. Some functions are exported, so it can be used to measure   *
 * normal time too, like a stopper.                                          *
 *****************************************************************************/

#include "FrameRegulation.h"

TID  FrameRegulatorThread; // Frame Regulation thread ID
ULONG ulFRPollStartTime;   // Start timestamp of actual frame
float flFRPollTime;        // Time (ms) to wait for a frame
int fFRShutdown;           // Indicator of shutdown request for FRAPI
HEV hevFRBlitNeededEvent;  // Event semaphore, posted by FrameRegulator, when it's time to tick!
ULONG ulInitTimeStamp;     // Timestamp, when the FrameRegulator has been initialized. Exported, so other threads can use it!
ULONG ulCurrTimeStamp;     // Current time, since boot of OS/2, in msecs. Exported, so other threads can use it!

// ----------- Frame Regulation -----------------------------

//---------------------------------------------------------------------
// Initialize_ulInitTimeStamp
//
// Initializes the ulInitTimeStamp variable, or in other words, queries
// the actual timestamp, and stores it in ulInitTimeStamp.
// The ulCurrTimeStamp is updated regularly, so counting the difference
// of these two, the time elapsed can be counted easily.
//---------------------------------------------------------------------

void Initialize_ulInitTimeStamp()
{
  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulInitTimeStamp, 4 );
}

//---------------------------------------------------------------------
// FrameRegulator
//
// This is the thread function for frame regulation.
// It frequently polls the actual time, and if it's time to do a tick,
// it posts a given semaphore.
//---------------------------------------------------------------------
VOID FrameRegulator(PVOID Param)
{
  ULONG ulCurrTime, ulPrevPollStartTime;
  float flNextFrameTime;

#ifdef FR_DEBUG_BUILD
  printf("Starting FrameRegulator thread\n");
  fflush(stdout);
#endif

  ulPrevPollStartTime = 0;
  flNextFrameTime += flFRPollTime;

  ulInitTimeStamp = ulFRPollStartTime; // Save timestamp of program initialization

  while (!fFRShutdown)
  {
    // Query the actual time of system
    DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulCurrTime, 4 );
    if (ulPrevPollStartTime != ulFRPollStartTime)
    { // The FPS setting has been changed on the fly!
      flNextFrameTime = ulPrevPollStartTime = ulFRPollStartTime;
    }
    // Calculate the next timestamp, when we will have to post a
    // semaphore.
    flNextFrameTime += flFRPollTime;
    // Wait for this time!
    while ((!fFRShutdown) &&
	   (ulCurrTime < flNextFrameTime) &&
	   (ulPrevPollStartTime == ulFRPollStartTime))
    {
      DosSleep(1);
      DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulCurrTime, 4 );
      ulCurrTimeStamp = ulCurrTime;
    }
    // Ok, time elapsed, post the semaphore!
    DosPostEventSem( hevFRBlitNeededEvent );
  }

#ifdef FR_DEBUG_BUILD
  printf("Halting FrameRegulator thread\n");
  fflush(stdout);
#endif

  _endthread();
}


//---------------------------------------------------------------------
// SetupFrameRegulation
//
// Queries the current timestamp, and calculates, how much to wait
// between the ticks. It can be called anytime, if the FPS should
// be changed.
//---------------------------------------------------------------------
void SetupFrameRegulation( float flFPS )
{
#ifdef FR_DEBUG_BUILD
  printf("SetupFrameRegulation( %f )\n", flFPS);
  fflush(stdout);
#endif
  if (flFPS==0.0) return;
  flFRPollTime = 1000.0 / flFPS;
  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulFRPollStartTime, 4 );
#ifdef FR_DEBUG_BUILD
  printf("  flFRPollTime: %f \n", flFRPollTime);
  fflush(stdout);
#endif
}

//---------------------------------------------------------------------
// InitializeFrameRegulation
//
// Starts frame regulation thread
//---------------------------------------------------------------------
int InitializeFrameRegulation(HEV hevSem)
{
#ifdef FR_DEBUG_BUILD
  printf("InitializeFrameRegulation()\n");
  fflush(stdout);
#endif

  fFRShutdown = 0;
  SetupFrameRegulation(1); // Default to 1 FPS
  hevFRBlitNeededEvent = hevSem; // Save semaphore handle to post
  FrameRegulatorThread = _beginthread( FrameRegulator,
				       NULL,
				       16384,
				       NULL);

  if (FrameRegulatorThread<1) return FALSE;
  // Burst the priority of regulator
  DosSetPriority(PRTYS_THREAD, PRTYC_FOREGROUNDSERVER, PRTYD_MAXIMUM, FrameRegulatorThread);
  return TRUE;
}

//---------------------------------------------------------------------
// UninitializeFrameRegulation
//
// Signals stop to frameregulator thread, and waits until it stops.
//---------------------------------------------------------------------
void UninitializeFrameRegulation()
{
#ifdef FR_DEBUG_BUILD
  printf("UninitializeFrameRegulation()\n");
  fflush(stdout);
#endif

  if (FrameRegulatorThread>=1)
  {
    fFRShutdown = 1;
    DosWaitThread(&FrameRegulatorThread, DCWW_WAIT);
  }
}
