/*****************************************************************************
 * Image support library                                                     *
 *                                                                           *
 * This library can be used to load and unpack images into BGR3 truecolor    *
 * format.                                                                   *
 * Currently only truecolor PCX files are supported.                         *
 *****************************************************************************/
#include "Images.h"

// We will keep track of all loaded images, so we can free them
// at uninit time, if needed.
typedef struct _tLoadedImageList
{
  pImage Image;
  void *Next;
} tLoadedImageList, *pLoadedImageList;

pLoadedImageList LoadedImageList = NULL;

// Definition of PCX header structure:
#pragma pack(push, 1)
typedef struct _tPCXHeader
{
  unsigned char Manufacturer;
  unsigned char Version;
  unsigned char Encoding;
  unsigned char BitsPerPixel;
  short int Window[4];
  short int HDpi;
  short int VDpi;
  unsigned char Colormap[3*16];
  unsigned char Reserved;
  unsigned char NPlanes;
  short int BytesPerLine;
  short int PaletteInfo;
  short int HscreenSize;
  short int VscreenSize;
  unsigned char Filler[54];
} tPCXHeader;
#pragma pack(pop)

//---------------------------------------------------------------------
// LoadPCX
//
// Loads a truecolor PCX, and returns a pImage to the uncompressed
// image. Only for 24bits PCX images!
//---------------------------------------------------------------------
pImage LoadPCX(char *filename)
{
  pImage result;
  FILE *PCXFile;
  tPCXHeader Header;
  long linesize;
  long count;
  unsigned char *linebuf, *imgypos;
  unsigned char *dst;
  long y, i;
  unsigned char b1, b2;
  unsigned char *filedata;
  long VirtFilePos;
  long filesize;

#ifdef IMAGES_DEBUG_BUILD
  printf("LoadPCX('%s')\n", filename);
  fflush(stdout);
#endif

  result=(pImage)malloc(sizeof(tImage));
  if (result)
  {

    if ((PCXFile=fopen(filename, "rb"))!=NULL)
    { // If we could open the file:
      // Read the file into memory!
      fseek(PCXFile, 0, SEEK_END);
      filesize=ftell(PCXFile);
      fseek(PCXFile, 0, SEEK_SET);
      
      filedata = (unsigned char *)malloc(filesize);
      if (filedata)
      {
        if (fread(filedata, 1, filesize, PCXFile)!=filesize)
	{ // Error!
          free(filedata);
          free(result); // Error!
          result=NULL;
          fclose(PCXFile);
          return result;
	};
	fclose(PCXFile);

      	// Ok, we have the whole file in memory, so
	// we can process it here!
	VirtFilePos=0;

	memcpy(&Header, filedata, sizeof(Header));
	VirtFilePos+=sizeof(Header);

	result->ulWidth=Header.Window[2]-Header.Window[0]+1;
	result->ulHeight=Header.Window[3]-Header.Window[1]+1;

	linesize=Header.NPlanes*Header.BytesPerLine;

	if ( (!(result->ulWidth)) || (!(result->ulHeight)) ||
	     (Header.BitsPerPixel!=8) || (Header.NPlanes!=3) )
	{ // This is not a 24bits PCX file, so we quit!
          free(filedata);
          free(result); // Error!
          result=NULL;
          return result;
	}
	// Get memory for image data (BGR)
	result->ulImageSize = result->ulWidth * result->ulHeight *3;
	result->pImageData = (unsigned char *) malloc(result->ulImageSize);
	if (!result->pImageData)
	{ // Could not allocate memory!
          free(filedata);
          free(result); // Error!
          result=NULL;
          return result;
	}

	imgypos = result->pImageData;
	linebuf = (unsigned char *)malloc(linesize+4);
	if (!linebuf)
	{
          free(result->pImageData);
          free(filedata);
          free(result); // Error!
          result=NULL;
          return result;
	}
	// Decode PCX planes!
      
	y=0;
	while (y<result->ulHeight)
	{
	  dst=linebuf;
	  count=0;
	  do
	  {
	    b1=filedata[VirtFilePos++]; // Get one byte!
	    if ((b1 & 0xC0) == 0xC0)              // Is it a lot of bytes?
	    {
	      b1=(b1 & 0x3f);             // Get number of bytes
	      b2=filedata[VirtFilePos++]; // and the color
	      memset(dst, b2, b1);
	      dst+=b1;
	      count+=b1;
	    } else
	    {                           // It's just one color value
	      *dst=b1;
	      dst++;
	      count++;
	    }
	  } while (count<linesize);
	  // Now we have one line decoded in linebuf.
	  // Let's store it in image! (in BGR format)

	  for (i=0; i<result->ulWidth; i++)
	  {
	    imgypos[i*3+2] = linebuf[i];                   //R
	    imgypos[i*3+1] = linebuf[(result->ulWidth)+i];   //G
            imgypos[i*3  ] = linebuf[2*(result->ulWidth)+i]; //B
	  }

	  y++;
          imgypos+=linesize;
	}

        free(linebuf);
	free(filedata);
	result->sFileName = (char *)malloc(strlen(filename)+1);
	strcpy(result->sFileName, filename);

      } else
      {
	free(result); // Error!
        result=NULL;
	fclose(PCXFile);
        return result;
      }
    } else
    { // If we could not open the file:
      free(result); // Error!
      result=NULL;
    }
  }

  if (result)
  { // Add to the list of loaded images
    pLoadedImageList NewImage = (pLoadedImageList) malloc(sizeof(tLoadedImageList));
    if (NewImage)
    {
      NewImage->Image = result;
      NewImage->Next = LoadedImageList;
      LoadedImageList = NewImage;
    }
  }
  return result;
}

//---------------------------------------------------------------------
// CleanupImage
//
// Cleans up a given image, and removes it from the list of loaded 
// images.
//---------------------------------------------------------------------
void CleanupImage(pImage Image)
{
  pLoadedImageList LoadedImage, prev;

  if (!Image) return;

#ifdef IMAGES_DEBUG_BUILD
  printf("CleanupImage()\n");
  fflush(stdout);
#endif

  // Search for the image in the loaded images list
  prev = NULL;
  LoadedImage = LoadedImageList;
  while ((LoadedImage) && (LoadedImage->Image != Image))
  {
    prev = LoadedImage;
    LoadedImage = LoadedImage->Next;
  }
  // If found, remove from the list!
  if (LoadedImage)
  {
    if (prev)
      prev->Next = LoadedImage->Next;
    else
      LoadedImageList = LoadedImage->Next;
    free(LoadedImage);
  }
  // And then, do the cleanup!
  free(Image->sFileName);
  free(Image->pImageData);
  free(Image);
}

//---------------------------------------------------------------------
// InitializeImages
//
// Initializes image support library.
//---------------------------------------------------------------------
void InitializeImages()
{
#ifdef IMAGES_DEBUG_BUILD
  printf("InitializeImages()\n");
  fflush(stdout);
#endif
  LoadedImageList = NULL;
}

//---------------------------------------------------------------------
// UninitializeImages
//
// Frees images that are not freed yet.
//---------------------------------------------------------------------
void UninitializeImages()
{
#ifdef IMAGES_DEBUG_BUILD
  printf("UninitializeImages()\n");
  fflush(stdout);
#endif

  while (LoadedImageList)
    CleanupImage(LoadedImageList->Image);
}
