/*****************************************************************************
 * OGG decoder module for the SoundFX library                                *
 *                                                                           *
 * This library is a wrapper for the integer-only OGG decoder, called TREMOR *
 *****************************************************************************/
#define INCL_TYPES
#define INCL_DOS
#define INCL_DOSERRORS
#include <os2.h>
#include <stdlib.h>
#include <stdio.h>

#include "oggidec.h"   // definitions for OGGIDEC.DLL exports
#include "OGGModule.h" // Module header file

static HMODULE hOGGDll = NULL;           // Handle for the DLL

t_ov_open_callbacks *ov_open_callbacks;  // Functions, used from the DLL
t_ov_test_callbacks *ov_test_callbacks;
t_ov_clear          *ov_clear;
t_ov_read           *ov_read;
t_ov_pcm_seek_page  *ov_pcm_seek_page;
t_ov_time_tell      *ov_time_tell;
t_ov_time_total     *ov_time_total;
t_ov_info           *ov_info;

// The OGG decoder uses callbacks for reading/seeking in a stream. This
// way, we can implement two kind of OGG loading:
// - File-streamed : read from file when data is needed
//   This requires small memory, but it might take a while to start a sound,
//   because the data has to be read from the HDD.
// - Memory-streamed : read the whole file into memory
//   Not memory-efficient, but fast access to compressed audio data.

//----- Default callbacks for file reading: -----------

size_t file_read(void *ptr, size_t size, size_t nmemb, void *datasource)
{
  ULONG ulActual;
#ifdef CALLBACK_DEBUG
  printf("file_read %p %d %d %p\n", ptr, size, nmemb, datasource); fflush(stdout);
#endif
  DosRead((HFILE)datasource, ptr, (size*nmemb), &ulActual);
  return (ulActual/size);
}

int file_seek(void *datasource, ogg_int64_t offset, int whence)
{
  ULONG ulMethod;
  ULONG ibActual;
#ifdef CALLBACK_DEBUG
  printf("file_seek %p %d %d\n", datasource, offset, whence); fflush(stdout);
#endif

  switch (whence)
  {
    case SEEK_SET:
      ulMethod = FILE_BEGIN;
      break;
    case SEEK_CUR:
      ulMethod = FILE_CURRENT;
      break;
    case SEEK_END:
      ulMethod = FILE_END;
      break;
    default:
      return 0;
  }
  DosSetFilePtr((HFILE)datasource, (LONG)(offset & 0xffffffff), ulMethod, &ibActual);
  return 0;
}

int file_close(void *datasource)
{
#ifdef CALLBACK_DEBUG
  printf("file_close %p\n", datasource); fflush(stdout);
#endif
  DosClose((HFILE)datasource);
  return 0;
}

long file_tell(void *datasource)
{
  ULONG ibActual;
#ifdef CALLBACK_DEBUG
  printf("file_tell %p\n", datasource); fflush(stdout);
#endif
  DosSetFilePtr((HFILE)datasource, 0, FILE_CURRENT, &ibActual);
  return ibActual;
}

//----- Default callbacks for memory reading: -----------

// Structure describing the memory-mapped file
typedef struct {
  char *pFileData;
  long lFileSize;
  long lFilePos;
} MEM_DESC, *PMEM_DESC;

size_t mem_read(void *ptr, size_t size, size_t nmemb, void *datasource)
{
  long lToRead = size * nmemb;
  PMEM_DESC MemDesc = (PMEM_DESC) datasource;

#ifdef CALLBACK_DEBUG
  printf("mem_read %p %d %d %p\n", ptr, size, nmemb, datasource); fflush(stdout);
#endif
  if (!MemDesc) return 0;

  if (lToRead>(MemDesc->lFileSize - MemDesc->lFilePos)) lToRead = MemDesc->lFileSize - MemDesc->lFilePos;

  if (lToRead>0)
  {
    memcpy(ptr, (MemDesc->pFileData) + MemDesc->lFilePos, lToRead);
    MemDesc->lFilePos+=lToRead;
  }
  return lToRead;
}

int mem_seek(void *datasource, ogg_int64_t offset, int whence)
{
  ULONG ulMethod;
  ULONG ibActual;
  PMEM_DESC MemDesc = (PMEM_DESC) datasource;

#ifdef CALLBACK_DEBUG
  printf("mem_seek %p %d %d\n", datasource, offset, whence); fflush(stdout);
#endif

  if (!MemDesc) return 0;

  switch (whence)
  {
    case SEEK_SET:
      MemDesc->lFilePos = (LONG)(offset & 0xffffffff);
      break;
    case SEEK_CUR:
      MemDesc->lFilePos += (LONG)(offset & 0xffffffff);
      break;
    case SEEK_END:
      MemDesc->lFilePos = MemDesc->lFileSize + (LONG)(offset & 0xffffffff);
      break;
    default:
      return 0;
  }

  if (MemDesc->lFilePos>=MemDesc->lFileSize) MemDesc->lFilePos = MemDesc->lFileSize-1;
  if (MemDesc->lFilePos<0) MemDesc->lFilePos = 0;

  return 0;
}

int mem_close(void *datasource)
{
  PMEM_DESC MemDesc = (PMEM_DESC) datasource;
#ifdef CALLBACK_DEBUG
  printf("mem_close %p\n", datasource); fflush(stdout);
#endif
  if (!MemDesc) return 0;
  free(MemDesc->pFileData);
  free(MemDesc);
  return 0;
}

long mem_tell(void *datasource)
{
  PMEM_DESC MemDesc = (PMEM_DESC) datasource;
#ifdef CALLBACK_DEBUG
  printf("mem_tell %p\n", datasource); fflush(stdout);
#endif
  if (!MemDesc) return 0;

  return MemDesc->lFilePos;
}

ov_callbacks file_ov_callbacks = {file_read, file_seek, file_close, file_tell};
ov_callbacks memory_ov_callbacks = {mem_read, mem_seek, mem_close, mem_tell};

// ------ OGG Decoder module functions ------------

// OGGUninit
//  Unloads OGG dll from memory
int OGGUninit()
{
  ULONG rc;
  if (hOGGDll)
  {
    rc = DosFreeModule(hOGGDll);
    if (rc==NO_ERROR) hOGGDll = NULL;
  }
  return (!hOGGDll);
}

// OGGInit
//  Loads OGG dll, and gets procedure entry points
int OGGInit()
{
  ULONG rc;

  if (hOGGDll)
    return 1;

  // Load oggidec.dll
  rc = DosLoadModule(NULL, 0, "oggidec.dll", &hOGGDll);
  if (rc!=NO_ERROR) return 0;

  // Query used procedure addresses
  rc = DosQueryProcAddr( hOGGDll, 0, "ov_open_callbacks", (PFN *)&ov_open_callbacks );
  if (rc!=NO_ERROR)
  {
    OGGUninit();
    return 0;
  }
  rc = DosQueryProcAddr( hOGGDll, 0, "ov_clear", (PFN *)&ov_clear );
  if (rc!=NO_ERROR)
  {
    OGGUninit();
    return 0;
  }
  rc = DosQueryProcAddr( hOGGDll, 0, "ov_read", (PFN *)&ov_read );
  if (rc!=NO_ERROR)
  {
    OGGUninit();
    return 0;
  }
  rc = DosQueryProcAddr( hOGGDll, 0, "ov_test_callbacks", (PFN *)&ov_test_callbacks );
  if (rc!=NO_ERROR)
  {
    OGGUninit();
    return 0;
  }
  rc = DosQueryProcAddr( hOGGDll, 0, "ov_pcm_seek_page", (PFN *)&ov_pcm_seek_page );
  if (rc!=NO_ERROR)
  {
    OGGUninit();
    return 0;
  }
  rc = DosQueryProcAddr( hOGGDll, 0, "ov_time_tell", (PFN *)&ov_time_tell );
  if (rc!=NO_ERROR)
  {
    OGGUninit();
    return 0;
  }
  rc = DosQueryProcAddr( hOGGDll, 0, "ov_time_total", (PFN *)&ov_time_total );
  if (rc!=NO_ERROR)
  {
    OGGUninit();
    return 0;
  }
  rc = DosQueryProcAddr( hOGGDll, 0, "ov_info", (PFN *)&ov_info );
  if (rc!=NO_ERROR)
  {
    OGGUninit();
    return 0;
  }
  return 1;
}

// OGGTest
//  Checks if a file is an OGG file or not.
int OGGTest(char *filename)
{
  HFILE hFile;
  ULONG rc;
  ULONG ulAction;
  int irc;
  OggVorbis_File vf;

  if (!hOGGDll) return 0;

  // Test if a given file is an OGG file or not:

  // First try to open the file!
  rc = DosOpen(filename, &hFile, &ulAction, 0, 0,
	       OPEN_ACTION_OPEN_IF_EXISTS | OPEN_ACTION_FAIL_IF_NEW,
               OPEN_FLAGS_NOINHERIT | OPEN_SHARE_DENYNONE | OPEN_ACCESS_READONLY, 0);

  // Return error if it does not exist.
  if (rc!=NO_ERROR) return 0;

  // Ok, file exist and opened, let the ogg library test it!
  irc = ov_test_callbacks((void *)hFile, &vf, NULL, 0, file_ov_callbacks);
  if (irc==0)
  {
    // Ok, this is an OGG file!
    // Clean up this ogg file (it will also close the file!)
    ov_clear(&vf);
    return 1;
  }
  else
  {
    // No, the file is not recognized to be an OGG file!
    // We should close the file, and return failure!
    DosClose(hFile);
    return 0;
  }
}

OGGHANDLE OGGOpen(char *filename, int iOpenMode)
{
  OggVorbis_File *result = (OggVorbis_File *)malloc(sizeof(OggVorbis_File));
  HFILE hFile;
  ULONG rc;
  ULONG ulAction;

  if (!result) return result; // Could not allocate memory!

  // First try to open the file!
  rc = DosOpen(filename, &hFile, &ulAction, 0, 0,
	       OPEN_ACTION_OPEN_IF_EXISTS | OPEN_ACTION_FAIL_IF_NEW,
               OPEN_FLAGS_NOINHERIT | OPEN_SHARE_DENYNONE | OPEN_ACCESS_READONLY, 0);

  // Return error if it does not exist.
  if (rc!=NO_ERROR) return 0;

  // Ok, file exist and opened, let the ogg library test it!
  if (iOpenMode == OPEN_OGG__FILESTREAM)
    rc = ov_open_callbacks((void *)hFile, result, NULL, 0, file_ov_callbacks);
  else
  if (iOpenMode == OPEN_OGG__MEMORYSTREAM)
  {
    PMEM_DESC MemFile = (PMEM_DESC) malloc(sizeof(MEM_DESC));
    if (!MemFile) rc = 1; // Indicate error!
    else
    { // Allocate memory for the file!
      FILESTATUS FileStatus;

      rc = DosQueryFileInfo(hFile, FIL_STANDARD, &FileStatus, sizeof(FileStatus));
      if (rc==NO_ERROR)
      { // Ok, we have filesize!
        // Now allocate memory enough to load the file into it!
        MemFile->lFileSize = FileStatus.cbFile;
        MemFile->lFilePos = 0;
        MemFile->pFileData = (char *) malloc(MemFile->lFileSize);
        if (MemFile->pFileData)
        {
          ULONG ulreaded;
          // All we have to do here to load the whole file into memory!
          rc = DosRead(hFile, MemFile->pFileData, MemFile->lFileSize, &ulreaded);
          if ((rc==NO_ERROR) && (ulreaded = MemFile->lFileSize))
          {
            // Ok, we could read the whole file into the memory!
            // Now let the OGG thingie do its stuff!
            rc = ov_open_callbacks((void *)MemFile, result, NULL, 0, memory_ov_callbacks);
          } else
          { // Error while reading the file!
            free(MemFile->pFileData);
            free(MemFile);
            rc = 1; // Indicate error!
          }
        } else
        { // Could not allocate memory to load the whole file into memory!
          free(MemFile);
          rc = 1; // Indicate error!
        }
      } else
      { // Could not allocate memory for MemFile descriptor!
        free(MemFile);
        rc = 1; // Indicate error!
      }
    }
  }
  else
    rc = 1; // Indicate error! (Invalid open mode!)

  if (rc!=0)
  {
    // The file is not recognized to be an OGG file!
    // We should close the file, and return failure!
    DosClose(hFile);
    free(result); result = NULL;
  } else
  if (iOpenMode == OPEN_OGG__MEMORYSTREAM)
  { // The memory stream has been loaded into memory, we can close the file!
    DosClose(hFile);
  }
  return result;
}

int OGGClose(OGGHANDLE Handle)
{
  if (!Handle) return 0;

  ov_clear((OggVorbis_File *)Handle);
  free(Handle);
  return 1;
}

int OGGSeek(OGGHANDLE Handle, long lPos)
{
  int rc;
  if (!Handle) return 0;

  rc = ov_pcm_seek_page((OggVorbis_File *)Handle, lPos);
  return (rc==0);
}

int OGGGetSampleRate(OGGHANDLE Handle)
{
  vorbis_info *vi;
  if (!Handle) return 0;

  vi = ov_info((OggVorbis_File *)Handle, -1); // for current stream
  if (vi==NULL) return 0;
  return (vi->rate);
}

int OGGGetChannels(OGGHANDLE Handle)
{
  vorbis_info *vi;
  if (!Handle) return 0;

  vi = ov_info((OggVorbis_File *)Handle, -1); // for current stream
  if (vi==NULL) return 0;
  return (vi->channels);
}

int OGGDecode(OGGHANDLE Handle, int iMode, unsigned char *pBuffer, unsigned long ulBufSize)
{
  int rc;
  int bitstream;
  int readed;
  OggVorbis_File *vf = Handle;

  if (!Handle) return 0;

  readed = 0;
  rc = 0;
  while ((rc!=OV_HOLE) && (rc!=OV_EBADLINK) && (readed<ulBufSize))
  {
    rc = ov_read((OggVorbis_File *)Handle, pBuffer, ulBufSize-readed, &bitstream);
    if (rc == 0)
    {
      // reached end of file/stream!
      if (iMode == DECODE_OGG__LOOPED)
      {
        // Loop, if necessary!
        ov_pcm_seek_page((OggVorbis_File *)Handle, 0);
      } else
      {
        // Not looped, fill remaining space with zeros!
        memset(pBuffer, 0, ulBufSize - readed);
        break;
      }
    } else
    if ((rc!=OV_HOLE) && (rc!=OV_EBADLINK))
    { // Readed some bytes!
      readed += rc;
      pBuffer += rc;
    }
  } // end of while
  // Return error, if there was any.
  if ((rc==0) || (rc==OV_HOLE) || (rc==OV_EBADLINK)) return 0;

  return 1;
}

__int64 OGGTellPos(OGGHANDLE Handle)
{
  return ov_time_tell((OggVorbis_File *)Handle);
}

__int64 OGGTellLength(OGGHANDLE Handle)
{
  OggVorbis_File *vf = Handle;
  if (!Handle) return -1;

  return ov_time_total((OggVorbis_File *)Handle, -1);
}

