/*****************************************************************************
 * Sprite support library                                                    *
 *                                                                           *
 * Using this library, one can use "Sprites", transparent and translucent    *
 * small images in his/her program.                                          *
 * The library uses the Images libary to load the images, from where the     *
 * sprite data is extracted.
 *****************************************************************************/
#include "Sprites.h"
#include <math.h>    // We're using floats too

// Some internal structures, that can be hidden from the outer world:

// When a sprite is put on to the output buffer, the data under it will be
// saved, so it is possible later to simply hide the sprite by redrawing
// the saved stuff into the saved position.
// These backgrounds will be put into a list of sprite backgrounds, which
// is used by RemoveDrawnSprites() to redraw all the backgrounds, in reverse
// order. (thus, restoring the screen in a quick way...)
typedef struct _tSpriteBgndList
{
  // What was under the sprite before displaying:
  unsigned char *pBackground;
  // And where it was
  unsigned long ulBgndXPos;
  unsigned long ulBgndYPos;
  // And how big it was
  unsigned long ulBgndXSize;
  unsigned long ulBgndYSize;

  void *Next;
} tSpriteBgndList, *pSpriteBgndList;

// List of loaded sprites, so we will be able to unload them at uninit time
typedef struct _tLoadedSpriteList
{
  pSprite Sprite;
  void *Next;
} tLoadedSpriteList, *pLoadedSpriteList;

pLoadedSpriteList LoadedSpriteList = NULL; // List of all the loaded sprites
pSpriteBgndList SpriteBgndList = NULL;     // List of the sprite backgrounds

// ------------  Functions ---------------

//---------------------------------------------------------------------
// AnimateSprite
//
// Animates a given sprite, using the animation properties stored in
// the sprite.
//---------------------------------------------------------------------
void AnimateSprite(pSprite Sprite)
{
  if (!Sprite) return;
  Sprite->flActFrame+=Sprite->flAnimSpeed;
  if (Sprite->flActFrame >= Sprite->ulNumOfFrames)
  {
    while (Sprite->flActFrame >= Sprite->ulNumOfFrames) Sprite->flActFrame-=Sprite->ulNumOfFrames;
    Sprite->flActFrame += Sprite->ulRestartFrame;
  }
  Sprite->ulActFrame=Sprite->flActFrame;
}

//---------------------------------------------------------------------
// RemoveDrawnSprite
//
// Draws the background of a sprite, so the sprite will be removed 
// from the screen.
// The background to be drawn is given as its 3rd parameter.
//---------------------------------------------------------------------
void RemoveDrawnSprite(unsigned char *pBuffer, unsigned long ulDstBufWidth, pSpriteBgndList SpriteBgnd)
{
  unsigned long xsize, ysize, dstAdd;
  unsigned char *src, *dst;
  // We should display the background of the sprite, if there is any
  if ((!pBuffer) || (!SpriteBgnd) || (!SpriteBgnd->pBackground)) return;

  xsize = SpriteBgnd->ulBgndXSize*3; // RGB
  ysize = SpriteBgnd->ulBgndYSize;
  src = SpriteBgnd->pBackground;
  dst = pBuffer + (SpriteBgnd->ulBgndYPos * ulDstBufWidth + SpriteBgnd->ulBgndXPos) * 3;
  dstAdd = ulDstBufWidth*3 - xsize;
  _asm
  {
    push eax
    push ecx
    push esi
    push edi

    mov esi, src
    mov edi, dst
    mov eax, ysize
  CopyMoreRS:
    mov ecx, xsize
    shr ecx, 2
    rep movsd
    mov ecx, xsize
    and ecx, 3
    rep movsb

    add edi, dstAdd
    dec eax
    jnz CopyMoreRS

    pop edi
    pop esi
    pop ecx
    pop eax
  };
  free(SpriteBgnd->pBackground);
  SpriteBgnd->pBackground=NULL;
}

//---------------------------------------------------------------------
// DrawSprite
//
// Saves the background into SpriteBackgroundList, and then draws the 
// sprite. It also cuts the sprite into a given viewport, and uses
// a given transparency.
//---------------------------------------------------------------------
void DrawSprite(unsigned char *pDstBuf, unsigned long ulDstBufWidth,
                long ulViewPortTLX, long ulViewPortTLY,
                long ulViewPortBRX, long ulViewPortBRY,
		long ulDstPosX, long ulDstPosY,
		pSprite Sprite,
		unsigned char bTransparency)
{
  long xsize, ysize, xsizet3;
  long xstart, ystart;
  unsigned long linesize;
  unsigned char *srcColor;
  unsigned char *srcMask;
  unsigned char *srcMaskCur;
  unsigned char *dst, *src;
  unsigned long ultemp;
  unsigned long dstAdd, srcAdd;
  pSpriteBgndList SpriteBgnd;

  if ((!pDstBuf) || (!Sprite)) return;

  ulDstPosX-=Sprite->ulHotSpotX;
  ulDstPosY-=Sprite->ulHotSpotY;

  // Cut for the viewport:
  xsize = Sprite->ulWidth;
  ysize = Sprite->ulHeight;
  xstart = ystart = 0;
  if (ulDstPosX+xsize >= ulViewPortBRX) xsize = ulViewPortBRX-ulDstPosX-1;
  if (ulDstPosY+ysize >= ulViewPortBRY) ysize = ulViewPortBRY-ulDstPosY-1;

  if (ulDstPosX<ulViewPortTLX)
  {
    xstart= labs(ulViewPortTLX-ulDstPosX);
    xsize-=xstart;
    ulDstPosX=ulViewPortTLX;
  }
  if (ulDstPosY<ulViewPortTLY)
  {
    ystart= labs(ulViewPortTLY-ulDstPosY);
    ysize-=ystart;
    ulDstPosY=ulViewPortTLY;
  }

  // Return, if there is nothing to show:
  if ((xsize<=0) || (ysize<=0) || (Sprite->ulActFrame > Sprite->ulNumOfFrames) ||
      (xstart>=Sprite->ulWidth) || (ystart>=Sprite->ulHeight)) return;

  // If there is something to put on the screen:
  // First save the background!
  SpriteBgnd = (pSpriteBgndList) malloc (sizeof (tSpriteBgndList));
  SpriteBgnd->ulBgndXPos = ulDstPosX;
  SpriteBgnd->ulBgndYPos = ulDstPosY;
  SpriteBgnd->ulBgndXSize = xsize;
  SpriteBgnd->ulBgndYSize = ysize;
  dst = SpriteBgnd->pBackground = (unsigned char *) malloc( xsize * ysize * 3 );
  src = pDstBuf + (ulDstPosY * ulDstBufWidth + ulDstPosX) * 3;
  srcAdd = (ulDstBufWidth - xsize) * 3;
  xsizet3 = xsize*3;
  _asm
  {
    push eax
    push ecx
    push esi
    push edi

    mov esi, src
    mov edi, dst
    mov eax, ysize
  CopyMoreDS:
    mov ecx, xsizet3
    shr ecx, 2
    rep movsd
    mov ecx, xsizet3
    and ecx, 3
    rep movsb

    add esi, srcAdd
    dec eax
    jnz CopyMoreDS

    pop edi
    pop esi
    pop ecx
    pop eax
  };
  // Now put it into the list of backgrounds!
  SpriteBgnd->Next = SpriteBgndList;
  SpriteBgndList = SpriteBgnd;

  // Show the sprite!
  linesize = xsize * 3;
  ultemp = (Sprite->ulActFrame * Sprite->ulOneFrameSize) + (ystart*Sprite->ulWidth + xstart) * 3;
  srcColor = Sprite -> pColorData + ultemp;
  srcMask = Sprite -> pColorMask + ultemp;
  dst = pDstBuf + (ulDstPosY * ulDstBufWidth + ulDstPosX) * 3;

  dstAdd = ulDstBufWidth*3 - linesize;
  srcAdd = Sprite->ulWidth*3 - linesize;

  _asm
  {
    push eax
    push ebx
    push ecx
    push edx
    push esi
    push edi

    mov esi, srcColor
    mov edi, dst

    mov ecx, ysize
  MoreLines:

    push ecx
    mov ecx, xsize
  MoreDataInThisLine:

    push ecx
    mov ebx, srcMask
    mov al, byte ptr [ebx]
    add ebx, 3
    mov srcMask, ebx
    mul byte ptr bTransparency
    mov cl, 255
    div cl
    mov bh, al
    mov bl, al
    neg bh
    dec bh

    lodsb
    mul bl
    mov dx, ax
    mov al, [edi]
    mul bh
    add ax,dx
    div cl
    stosb

    lodsb
    mul bl
    mov dx, ax
    mov al, [edi]
    mul bh
    add ax,dx
    div cl
    stosb

    lodsb
    mul bl
    mov dx, ax
    mov al, [edi]
    mul bh
    add ax,dx
    div cl
    stosb

    pop ecx                // Decrement X counter
    dec ecx
    jnz MoreDataInThisLine
    // End of one line, advance to next line!
    mov ecx, srcAdd
    add edi, dstAdd
    add esi, ecx
    add srcMask, ecx

    pop ecx                // Decrement Y counter
    dec ecx
    jnz MoreLines

    pop edi
    pop esi
    pop edx
    pop ecx
    pop ebx
    pop eax
  };
}

//---------------------------------------------------------------------
// RemoveDrawnSprites
//
// Goes through the list of saved sprite backgrounds, and puts them
// onto the screen, thus, removing the sprites.
//---------------------------------------------------------------------
void RemoveDrawnSprites(unsigned char *pBuffer, unsigned long ulDstBufWidth)
{
  pSpriteBgndList toremove;
  while (SpriteBgndList)
  {
    toremove = SpriteBgndList;
    RemoveDrawnSprite(pBuffer, ulDstBufWidth, toremove);
    SpriteBgndList = SpriteBgndList->Next;
    free(toremove);
  }
}

//---------------------------------------------------------------------
// LoadSprite
//
// Extracts sprite properties and data from the given image and sprite
// definition parameters.
//---------------------------------------------------------------------
pSprite LoadSprite(pImage Image, char *Params)
{
  pSprite result=NULL;
  char *remain;
  int rc;
  int c, y, dx, dy, mx, my;
  unsigned char *src, *dst, *srcM, *dstM;
  // Params is a string containing the following:
  // sprite dimensions, number of frames, speed of animation (float), and
  // top-left coordinates of frame color and mask positions
  // For an example, see the default skin file!

  remain = (char *) malloc(2048);
  
  if ((!Image) || (!Params) || (!remain)) return result;
  
  // Allocate memory for the sprite structure
  result=(pSprite)malloc(sizeof(tSprite));
  if (!result)
  {
    free(remain);
    return result;
  }
  
  // Zero out
  memset(result, 0, sizeof(tSprite));

  // Get the sprite parameters
  rc=sscanf(Params, "%d %d %d %d %d %d %f %[^\0]",
	    &(result->ulWidth),
	    &(result->ulHeight),
	    &(result->ulHotSpotX),
            &(result->ulHotSpotY),
            &(result->ulNumOfFrames),
            &(result->ulRestartFrame),
	    &(result->flAnimSpeed),
	    remain);

  if ((rc!=8) || (result->ulRestartFrame<=0) || (result->ulRestartFrame>result->ulNumOfFrames))
  {
    free(result); result=NULL;
    free(remain);
    return result;
  }
  // Normalize ulRestartFrame from [1..ulNumOfFrames] to [0..ulNumOfFrames-1]!
  result->ulRestartFrame--;

  // Ok, all we have to do now is to extract the color data and the mask data for
  // every coordinate-pair in 'remain' string!

  result->ulSize=result->ulWidth*result->ulHeight*3*result->ulNumOfFrames;
  result->pColorData=(unsigned char *)malloc(result->ulSize);
  if (!result->pColorData)
  {
    free(result); result=NULL;
    free(remain);
    return result;
  }
  result->pColorMask=(unsigned char *)malloc(result->ulSize);
  if (!result->pColorMask)
  {
    free(result->pColorData);
    free(result); result=NULL;
    free(remain);
    return result;
  }

  result->ulOneFrameSize = result->ulWidth * result->ulHeight * 3;
  for (c=0; c<result->ulNumOfFrames; c++)
  {
    rc=sscanf(remain, "%d %d %d %d %[^\0]",
	      &dx, &dy, // Color data position
	      &mx, &my, // Mask data position
              remain);
    if ((rc==5) || (rc==4))
    {
      // Copy color data and mask data from source image to
      // sprite definition structure!
      src=Image->pImageData + Image->ulWidth*dy*3 + dx*3;
      dst=result->pColorData + (c * result->ulOneFrameSize);
      srcM=Image->pImageData + Image->ulWidth*my*3 + mx*3;
      dstM=result->pColorMask + (c * result->ulOneFrameSize);
      for (y=0; y<result->ulHeight; y++)
      {
	memcpy(dst, src, result->ulWidth*3);
	dst+=result->ulWidth*3;
	src+=Image->ulWidth*3;
        memcpy(dstM, srcM, result->ulWidth*3);
	dstM+=result->ulWidth*3;
        srcM+=Image->ulWidth*3;
      }
    }
  }

  free(remain);

  if (result)
  { // Add to the list of loaded sprites
    pLoadedSpriteList NewSprite = (pLoadedSpriteList) malloc(sizeof(tLoadedSpriteList));
    if (NewSprite)
    {
      NewSprite->Sprite = result;
      NewSprite->Next = LoadedSpriteList;
      LoadedSpriteList = NewSprite;
    }
  }

  return result;
}

//---------------------------------------------------------------------
// CleanupSprite
//
// Cleans up a given sprite by freeing every resource allocated for it
//---------------------------------------------------------------------
void CleanupSprite(pSprite Sprite)
{
  pLoadedSpriteList LoadedSprite, prev;

  if (!Sprite) return;

#ifdef SPRITES_DEBUG_BUILD
  printf("CleanupSprite()\n");
  fflush(stdout);
#endif

  // Search for the sprite in the loaded sprites list
  prev = NULL;
  LoadedSprite = LoadedSpriteList;
  while ((LoadedSprite) && (LoadedSprite->Sprite != Sprite))
  {
    prev = LoadedSprite;
    LoadedSprite = LoadedSprite->Next;
  }
  // If found, remove from the list!
  if (LoadedSprite)
  {
    if (prev)
      prev->Next = LoadedSprite->Next;
    else
      LoadedSpriteList = LoadedSprite->Next;
    free(LoadedSprite);
  }
  // Free the sprite data!
  free(Sprite->pColorData);
  free(Sprite->pColorMask);
  free(Sprite);
}

//---------------------------------------------------------------------
// InitializeSprites
//
// Initialize sprite library
//---------------------------------------------------------------------
void InitializeSprites()
{
#ifdef SPRITES_DEBUG_BUILD
  printf("InitializeSprites()\n");
  fflush(stdout);
#endif

  SpriteBgndList = NULL;
  LoadedSpriteList = NULL;
}

//---------------------------------------------------------------------
// UninitializeSprites
//
// Free remaining resources, allocated by sprite library
//---------------------------------------------------------------------
void UninitializeSprites()
{
  pSpriteBgndList toremove;

#ifdef SPRITES_DEBUG_BUILD
  printf("UninitializeSprites()\n");
  fflush(stdout);
#endif

  // Free background list:
  while (SpriteBgndList)
  {
    toremove = SpriteBgndList;
    SpriteBgndList = SpriteBgndList->Next;
    free(toremove->pBackground);
    free(toremove);
  }
  // Free sprite list:
  while (LoadedSpriteList)
    CleanupSprite(LoadedSpriteList->Sprite);
}
